
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <pspkernel.h>
#include <pspdebug.h>
#include <pspdisplay.h>
#include <pspge.h>
#include <pspgu.h>
#include <pspctrl.h>
#include <pspaudio.h>
#include <pspaudiocodec.h>

#include "util.h"

PSP_MODULE_INFO("jam", 0, 1, 1);
PSP_MAIN_THREAD_ATTR(THREAD_ATTR_USER);
PSP_HEAP_SIZE_MAX();

#define ROWS 34
#define COLS 68

#define SAFE_TOP 3
#define SAFE_RIGHT 1
#define SAFE_BOTTOM 1
#define SAFE_LEFT 1
#define SAFE_ROWS (ROWS - SAFE_TOP - SAFE_BOTTOM)
#define SAFE_COLS (COLS - SAFE_LEFT - SAFE_RIGHT)

static int cursorX = SAFE_COLS / 2;
static int cursorY = SAFE_ROWS / 2;

static SceCtrlData pad;
static SceCtrlData lastPad;

static int frame;

static u32 sourceSize;
static u32 sourceLines;
static char **originalLines;
static char *source;
static char **lines;
static int numLines;
static int audioChannel;
static s16 *blipPcm;

int exitCallback(int arg1, int arg2, void *common)
{
    sceKernelExitGame();
    return 0;
}

int exitCallbackThread(SceSize args, void *argp)
{
    int cbid = sceKernelCreateCallback("exitCallback", exitCallback, 0);
    sceKernelRegisterExitCallback(cbid);
    sceKernelSleepThreadCB();
    return 0;
}

int setupExitCallbacks()
{
    int thid = sceKernelCreateThread("exitCallbackThread", exitCallbackThread, 0x11, 0xFA0, 0, 0);
    if (thid >= 0)
    {
        sceKernelStartThread(thid, 0, 0);
    }
    return 0;
}

static void drawChar(int x, int y, u32 bg, u32 fg, char c)
{
    pspDebugScreenSetXY(x, y);
    pspDebugScreenSetBackColor(bg);
    pspDebugScreenSetTextColor(fg);
    pspDebugScreenPrintf("%c", c);
}

static void blip()
{
    int size = 0x1000;
    if (!blipPcm)
    {
        blipPcm = (s16*)malloc(size * sizeof(s16));
        int i;
        float t;
        for (i = 0; i < size; ++i)
        {
            t = (float)i / 44100.f;
            blipPcm[i] = (s16)((float)INT16_MAX * sinf(t * 440.f * 8 * GU_PI));
        }
    }

    int ch = sceAudioChReserve(PSP_AUDIO_NEXT_CHANNEL, size, PSP_AUDIO_FORMAT_MONO);
    if (ch > 0)
    {
        sceAudioOutput(ch, PSP_AUDIO_VOLUME_MAX, blipPcm);
        sceAudioChRelease(ch);
    }
}

int main(int argc, char *argv[])
{
    setupExitCallbacks();
    SceUID fd = sceIoOpen("source.c", PSP_O_RDONLY, 0777);
    if (fd < 0)
    {
        return -1;
    }
    sourceSize = sceIoLseek32(fd, 0, PSP_SEEK_END);
    sceIoLseek32(fd, 0, PSP_SEEK_SET);
    source = (char*)malloc(sourceSize);
    sceIoRead(fd, source, sourceSize);
    sceIoClose(fd);

    int i, j;
    for (i = 0; i < sourceSize; ++i)
    {
        if (source[i] == '\n')
        {
            ++numLines;
        }
    }

    lines = (char**)malloc(sizeof(char*) * numLines);
    originalLines = (char**)malloc(sizeof(char*) * numLines);

    int left = 0, line = 0, len;
    for (i = 0; i < sourceSize; ++i)
    {
        if (source[i] == '\n' || i == sourceSize - 1)
        {
            len = max(2, sizeof(char) * (i - left));
            lines[line] = (char*)malloc(len);
            originalLines[line] = (char*)malloc(len);
            memcpy(lines[line], source + left, sizeof(char) * (i - left));
            lines[line][len - 1] = 0;
            memcpy(originalLines[line], lines[line], len);
            fprintf(stdout, "line %d: %s (%d)\n", line, lines[line], len);
            ++line;
            left = i + 1;
        }
    }

    fprintf(stdout, "lines %d\n", numLines);

    pspDebugScreenInit();
    pspDebugScreenClearLineDisable();

    u32 fbp0 = 0;
    u32 fbp1 = 272 * 512 * 4;
    u32 fbp = fbp0;
    u32 edram = (u32)sceGeEdramGetAddr();

	sceCtrlSetSamplingCycle(0);
	sceCtrlSetSamplingMode(PSP_CTRL_MODE_ANALOG);

    float analogX, analogY;
    int lastCursorX, lastCursorY;
    char selection;
    u32 newButtons;
    int lineLen;
    int totalChars;
    int bugChars;
    int charChanged = 0;
    float money = 0;
    int gameOver = 0;

    while (1)
    {
        charChanged = 0;
        lastPad = pad;
        sceCtrlPeekBufferPositive(&pad, 1);
        newButtons = pad.Buttons & ~lastPad.Buttons;
        analogX = clamp(((float)pad.Lx - 127.f) / 127.f, -1, 1);
        analogY = clamp(((float)pad.Ly - 127.f) / 127.f, -1, 1);

        lineLen = strlen(lines[min(cursorY, numLines - 1)]);

        if (!gameOver)
        {
            if (fabsf(analogX) > .5f)
            {
                cursorX = clamp(cursorX + sign(analogX), 0, lineLen - 1);
            }
            else if (fabsf(analogY) > .5f)
            {
                cursorY = clamp(cursorY + sign(analogY), 0, numLines - 1);
            }
            else if (newButtons & PSP_CTRL_UP)
            {
                cursorY = clamp(cursorY - 1, 0, numLines - 1);
            }
            else if (newButtons & PSP_CTRL_DOWN)
            {
                cursorY = clamp(cursorY + 1, 0, numLines - 1);
            }
            else if (newButtons & PSP_CTRL_LEFT)
            {
                cursorX = clamp(cursorX - 1, 0, lineLen - 1);
            }
            else if (newButtons & PSP_CTRL_RIGHT)
            {
                cursorX = clamp(cursorX + 1, 0, lineLen - 1);
            }
            if (newButtons & PSP_CTRL_RTRIGGER)
            {
                char *line = lines[cursorY];
                ++line[cursorX];
                charChanged = 1;
            }
            if (newButtons & PSP_CTRL_LTRIGGER)
            {
                char *line = lines[cursorY];
                --line[cursorX];
                charChanged = 1;
            }
        }

        if (charChanged && lines[cursorY][cursorX] == originalLines[cursorY][cursorX])
        {
            blip();
            money += 100;
        }

        if (rand() % 100 == 0 || (gameOver && rand() % 10 == 0))
        {
            int times = 1;
            if (money > 2000 && (rand() % 1000) == 0)
            {
                times = 10;
            }
            for (; times >= 0; --times)
            {
                char *line = lines[rand() % numLines];
                --line[rand() % strlen(line)];
            }
        }

        pspDebugScreenSetBackColor(0);
        pspDebugScreenSetTextColor(0xffffffff);

        pspDebugScreenClear();

        pspDebugScreenSetXY(0, 3);

        totalChars = 0;
        bugChars = 0;
        for (i = 0; i < numLines; ++i)
        {
            for (j = 0; j < strlen(lines[i]); ++j)
            {
                if (lines[i][j] != originalLines[i][j])
                {
                    ++bugChars;
                }
                ++totalChars;
                pspDebugScreenSetBackColor(i == cursorY && j == cursorX ? 0xffffffff : i == cursorY ? 0xff202020 : 0);
                pspDebugScreenSetTextColor(lines[i][j] != originalLines[i][j] ? 0xff0000ff : 0xffffffff);
                pspDebugScreenPrintf("%c", lines[i][j]);
            }
            if (j == 1)
            {
                pspDebugScreenPrintf(" ");
            }
            pspDebugScreenPrintf("\n");
        }

        if (totalChars == 0)
        {
            continue;
        }

        if ((int)((float)(totalChars - bugChars) / (float)totalChars * 100.f) < 95)
        {
            gameOver = 1;
        }

        pspDebugScreenSetXY(25, 1);
        pspDebugScreenSetBackColor(0);
        pspDebugScreenSetTextColor(0xff00ff00);
        pspDebugScreenPrintf("CODE COVERAGE: %3d%%", (int)((float)(totalChars - bugChars) / (float)totalChars * 100.f));

        pspDebugScreenSetXY(1, 1);
        pspDebugScreenSetTextColor(0xffffffff);
        pspDebugScreenPrintf("<-- DEC. CHAR");
        pspDebugScreenSetXY(55, 1);
        pspDebugScreenPrintf("INC. CHAR -->");

        pspDebugScreenSetTextColor(0xff0000ff);
        if (gameOver)
        {
            pspDebugScreenSetXY(1, 24);
            pspDebugScreenPrintf("GAME OVER :(");
        }
        pspDebugScreenSetTextColor(0xff00ff00);
        pspDebugScreenSetXY(1, 26);
        pspDebugScreenPrintf("MONEY: $%4.2f", money);
        pspDebugScreenSetTextColor(0xff00ffff);
        pspDebugScreenSetXY(1, 29);
        pspDebugScreenPrintf("%c listen John, don't let code coverage fall below 95%% again.", 2);
        pspDebugScreenSetXY(1, 30);
        pspDebugScreenPrintf("Those pesky bugs are really eating into the margins.");
        pspDebugScreenSetXY(1, 31);
        pspDebugScreenPrintf("I am not paying you $100 a minute for nothing... not to mention the bonuses... Now get to work.");

        sceDisplayWaitVblankStart();
        sceDisplaySetFrameBuf((void*)(edram + fbp), 512, PSP_DISPLAY_PIXEL_FORMAT_8888, 0);
        fbp = fbp == fbp0 ? fbp1 : fbp0;
        pspDebugScreenSetBase((u32*)(edram + fbp));

        ++frame;
        if (!gameOver)
        {
            money += 1.f / 30.f / 60.f * 100.f;
        }

        lastCursorX = cursorX;
        lastCursorY = cursorY;
    }

    return 0;
}